@@ -1,30 +1,37 @@
 r"""Parameterizable probability distributions."""
 
 __all__ = [
-    'NormalizingFlow',
-    'Joint',
-    'Mixture',
-    'GeneralizedNormal',
-    'DiagNormal',
-    'BoxUniform',
-    'TransformedUniform',
-    'Truncated',
-    'Sort',
-    'TopK',
-    'Minimum',
-    'Maximum',
+    "BoxUniform",
+    "DiagNormal",
+    "GeneralizedNormal",
+    "Joint",
+    "Maximum",
+    "Minimum",
+    "Mixture",
+    "NormalizingFlow",
+    "Sort",
+    "TopK",
+    "TransformedUniform",
+    "Truncated",
 ]
 
 import math
 import torch
 
 from textwrap import indent
-from torch import Tensor, Size
-from torch.distributions import *
-from torch.distributions import constraints
+from torch import Size, Tensor
+from torch.distributions import *  # noqa: F403
+from torch.distributions import (
+    Categorical,
+    Distribution,
+    Independent,
+    Normal,
+    Transform,
+    Uniform,
+    constraints,
+)
 from torch.distributions.utils import _sum_rightmost
-from typing import *
-
+from typing import Tuple
 
 Distribution._validate_args = False
 Distribution.arg_constraints = {}
@@ -52,13 +59,13 @@ class NormalizingFlow(Distribution):
         | https://arxiv.org/abs/1912.02762
 
     Arguments:
-        transforms: A transformation :math:`f`.
+        transform: A transformation :math:`f`.
         base: A base distribution :math:`p(Z)`.
 
     Example:
         >>> d = NormalizingFlow(ExpTransform(), Gamma(2.0, 1.0))
         >>> d.sample()
-        tensor(1.1316)
+        tensor(1.5157)
     """
 
     has_rsample = True
@@ -81,12 +88,12 @@ class NormalizingFlow(Distribution):
 
     def __repr__(self) -> str:
         lines = [
-            f'(transform): {self.transform}',
-            f'(base): {self.base}',
+            f"(transform): {self.transform}",
+            f"(base): {self.base}",
         ]
-        lines = indent('\n'.join(lines), '  ')
+        lines = indent("\n".join(lines), "  ")
 
-        return self.__class__.__name__ + '(\n' + lines + '\n)'
+        return self.__class__.__name__ + "(\n" + lines + "\n)"
 
     @property
     def batch_shape(self) -> Size:
@@ -96,7 +103,7 @@ class NormalizingFlow(Distribution):
     def event_shape(self) -> Size:
         return self.transform.inverse_shape(self.base.event_shape)
 
-    def expand(self, batch_shape: Size, new: Distribution = None):
+    def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
         new = self._get_checked_instance(NormalizingFlow, new)
         new.transform = self.transform
         new.base = self.base.expand(batch_shape)
@@ -146,7 +153,7 @@ class Joint(Distribution):
         >>> d.event_shape
         torch.Size([2])
         >>> d.sample()
-        tensor([ 0.8969, -2.6717])
+        tensor([0.4963, 0.2072])
     """
 
     has_rsample = True
@@ -160,9 +167,9 @@ class Joint(Distribution):
 
     def __repr__(self) -> str:
         lines = map(repr, self.marginals)
-        lines = indent('\n'.join(lines), '  ')
+        lines = indent("\n".join(lines), "  ")
 
-        return self.__class__.__name__ + '(\n' + lines + '\n)'
+        return self.__class__.__name__ + "(\n" + lines + "\n)"
 
     @property
     def batch_shape(self) -> Size:
@@ -224,7 +231,7 @@ class Mixture(Distribution):
         >>> d.event_shape
         torch.Size([])
         >>> d.sample()
-        tensor(2.8732)
+        tensor(-1.6920)
     """
 
     has_rsample = False
@@ -239,7 +246,7 @@ class Mixture(Distribution):
         self.logits = logits
 
     def __repr__(self) -> str:
-        return f'{self.__class__.__name__}({self.base})'
+        return f"{self.__class__.__name__}({self.base})"
 
     @property
     def batch_shape(self) -> Size:
@@ -294,10 +301,10 @@ class GeneralizedNormal(Distribution):
     Example:
         >>> d = GeneralizedNormal(2.0)
         >>> d.sample()
-        tensor(0.7480)
+        tensor(-0.0281)
     """
 
-    arg_constraints = {'beta': constraints.positive}
+    arg_constraints = {"beta": constraints.positive}
     support = constraints.real
     has_rsample = True
 
@@ -319,11 +326,7 @@ class GeneralizedNormal(Distribution):
         return new
 
     def log_prob(self, x: Tensor) -> Tensor:
-        return (
-            torch.log(self.beta / 2)
-            - torch.lgamma(1 / self.beta)
-            - abs(x) ** self.beta
-        )
+        return torch.log(self.beta / 2) - torch.lgamma(1 / self.beta) - abs(x) ** self.beta
 
     def rsample(self, shape: Size = ()) -> Tensor:
         beta = self.beta.expand(shape + self.beta.shape)
@@ -347,14 +350,14 @@ class DiagNormal(Independent):
         >>> d.event_shape
         torch.Size([3])
         >>> d.sample()
-        tensor([ 0.7304, -0.1976, -1.7591])
+        tensor([ 1.5410, -0.2934, -2.1788])
     """
 
     def __init__(self, loc: Tensor, scale: Tensor, ndims: int = 1):
         super().__init__(Normal(torch.as_tensor(loc), torch.as_tensor(scale)), ndims)
 
     def __repr__(self) -> str:
-        return 'Diag' + repr(self.base_dist)
+        return "Diag" + repr(self.base_dist)
 
     def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
         new = self._get_checked_instance(DiagNormal, new)
@@ -380,14 +383,14 @@ class BoxUniform(Independent):
         >>> d.event_shape
         torch.Size([3])
         >>> d.sample()
-        tensor([ 0.1859, -0.9698,  0.0665])
+        tensor([-0.0075,  0.5364, -0.8230])
     """
 
     def __init__(self, lower: Tensor, upper: Tensor, ndims: int = 1):
         super().__init__(Uniform(torch.as_tensor(lower), torch.as_tensor(upper)), ndims)
 
     def __repr__(self) -> str:
-        return 'Box' + repr(self.base_dist)
+        return "Box" + repr(self.base_dist)
 
     def expand(self, batch_shape: Size, new: Distribution = None) -> Distribution:
         new = self._get_checked_instance(BoxUniform, new)
@@ -413,7 +416,7 @@ class TransformedUniform(NormalizingFlow):
     Example:
         >>> d = TransformedUniform(ExpTransform(), -1.0, 1.0)
         >>> d.sample()
-        tensor(0.5594)
+        tensor(0.4281)
     """
 
     def __init__(self, f: Transform, lower: Tensor, upper: Tensor):
@@ -442,7 +445,7 @@ class Truncated(Distribution):
     Example:
         >>> d = Truncated(Normal(0.0, 1.0), 1.0, 2.0)
         >>> d.sample()
-        tensor(1.2573)
+        tensor(1.3333)
     """
 
     has_rsample = True
@@ -450,8 +453,8 @@ class Truncated(Distribution):
     def __init__(
         self,
         base: Distribution,
-        lower: Tensor = float('-inf'),
-        upper: Tensor = float('+inf'),
+        lower: Tensor = float("-inf"),
+        upper: Tensor = float("+inf"),
     ):
         super().__init__()
 
@@ -461,7 +464,7 @@ class Truncated(Distribution):
         self.uniform = Uniform(base.cdf(lower), base.cdf(upper))
 
     def __repr__(self) -> str:
-        return f'{self.__class__.__name__}({self.base})'
+        return f"{self.__class__.__name__}({self.base})"
 
     @property
     def batch_shape(self) -> Size:
@@ -483,7 +486,7 @@ class Truncated(Distribution):
         return self.uniform.log_prob(self.base.cdf(x)) + self.base.log_prob(x)
 
     def rsample(self, shape: Size = ()) -> Tensor:
-        return self.base.icdf(torch.clip(self.uniform.rsample(shape), 1e-6, 1 - 1e-6))
+        return self.base.icdf((1 - 2e-6) * self.uniform.rsample(shape) + 1e-6)
 
 
 class Sort(Distribution):
@@ -506,7 +509,7 @@ class Sort(Distribution):
         >>> d.event_shape
         torch.Size([3])
         >>> d.sample()
-        tensor([-1.4434, -0.3861,  0.2439])
+        tensor([-2.1788, -0.2934,  1.5410])
     """
 
     def __init__(
@@ -525,7 +528,7 @@ class Sort(Distribution):
         self.log_fact = math.log(math.factorial(n))
 
     def __repr__(self) -> str:
-        return f'{self.__class__.__name__}({self.base}, {self.n})'
+        return f"{self.__class__.__name__}({self.base}, {self.n})"
 
     @property
     def batch_shape(self) -> Size:
@@ -559,7 +562,7 @@ class Sort(Distribution):
         return ordered.log() + self.log_fact + self.base.log_prob(x).sum(dim=0)
 
     def sample(self, shape: Size = ()) -> Tensor:
-        x = torch.movedim(self.base.sample((self.n,) + shape), 0, -1)
+        x = torch.movedim(self.base.sample((self.n, *shape)), 0, -1)
         x = torch.sort(x, dim=-1, descending=self.descending).values
 
         return x
@@ -587,7 +590,7 @@ class TopK(Sort):
         >>> d.event_shape
         torch.Size([2])
         >>> d.sample()
-        tensor([-0.2167,  0.6739])
+        tensor([-2.1788, -0.2934])
     """
 
     def __init__(
@@ -605,7 +608,7 @@ class TopK(Sort):
         self.log_fact = self.log_fact - math.log(math.factorial(n - k))
 
     def __repr__(self) -> str:
-        return f'{self.__class__.__name__}({self.base}, {self.k}, {self.n})'
+        return f"{self.__class__.__name__}({self.base}, {self.k}, {self.n})"
 
     @property
     def event_shape(self) -> Size:
@@ -643,7 +646,7 @@ class Minimum(TopK):
         >>> d.event_shape
         torch.Size([])
         >>> d.sample()
-        tensor(-1.7552)
+        tensor(-2.1788)
     """
 
     def __init__(self, base: Distribution, n: int = 2):
@@ -684,7 +687,7 @@ class Maximum(Minimum):
         >>> d.event_shape
         torch.Size([])
         >>> d.sample()
-        tensor(1.1644)
+        tensor(1.5410)
     """
 
     def __init__(self, base: Distribution, n: int = 2):
