@@ -1,13 +1,25 @@
 r"""Neural networks, layers and modules."""
 
-__all__ = ['MLP', 'MaskedMLP', 'MonotonicMLP']
+__all__ = ["MLP", "Linear", "MaskedMLP", "MonotonicMLP"]
 
 import torch
 import torch.nn as nn
 import torch.nn.functional as F
 
-from torch import Tensor, BoolTensor
-from typing import *
+from torch import BoolTensor, Tensor
+from typing import Callable, Iterable, Sequence, Union
+
+
+def linear(x: Tensor, W: Tensor, b: Tensor = None) -> Tensor:
+    if W.dim() == 2:
+        return F.linear(x, W, b)
+    else:
+        x = torch.einsum("...ij,...j->...i", W, x)
+
+    if b is None:
+        return x
+    else:
+        return x + b
 
 
 class LayerNorm(nn.Module):
@@ -27,7 +39,7 @@ class LayerNorm(nn.Module):
     def __init__(self, dim: Union[int, Iterable[int]] = -1, eps: float = 1e-5):
         super().__init__()
 
-        self.dim = dim if type(dim) is int else tuple(dim)
+        self.dim = dim if isinstance(dim, int) else tuple(dim)
         self.eps = eps
 
     def forward(self, x: Tensor) -> Tensor:
@@ -36,6 +48,77 @@ class LayerNorm(nn.Module):
         return (x - mean) / (variance + self.eps).sqrt()
 
 
+class Linear(nn.Module):
+    r"""Creates a linear layer.
+
+    .. math:: y = x W^T + b
+
+    If the :py:`stack` argument is provided, the module creates a stack of
+    independent linear operators that are applied to a stack of input vectors.
+
+    Arguments:
+        in_features: The number of input features :math:`C`.
+        out_features: The number of output features :math:`C'`.
+        bias: Whether the layer learns an additive bias :math:`b` or not.
+        stack: The number of stacked operators :math:`S`.
+    """
+
+    def __init__(
+        self,
+        in_features: int,
+        out_features: int,
+        bias: bool = True,
+        stack: int = None,
+    ):
+        super().__init__()
+
+        shape = () if stack is None else (stack,)
+
+        self.weight = nn.Parameter(torch.empty(*shape, out_features, in_features))
+
+        if bias:
+            self.bias = nn.Parameter(torch.empty(*shape, out_features))
+        else:
+            self.bias = None
+
+        self.reset_parameters()
+
+        self.in_features = in_features
+        self.out_features = out_features
+
+    def reset_parameters(self):
+        bound = 1 / self.weight.shape[-1] ** 0.5
+
+        nn.init.uniform_(self.weight, -bound, bound)
+
+        if self.bias is not None:
+            nn.init.uniform_(self.bias, -bound, bound)
+
+    def extra_repr(self) -> str:
+        if self.weight.dim() == 2:
+            stack, fout, fin = (None, *self.weight.shape)
+        else:
+            stack, fout, fin = self.weight.shape
+
+        bias = self.bias is not None
+
+        if stack is None:
+            return f"in_features={fin}, out_features={fout}, bias={bias}"
+        else:
+            return f"in_features={fin}, out_features={fout}, bias={bias}, stack={stack}"
+
+    def forward(self, x: Tensor) -> Tensor:
+        r"""
+        Arguments:
+            x: The input tensor :math:`x`, with shape :math:`(*, C)` or :math:`(*, S, C)`.
+
+        Returns:
+            The output tensor :math:`y`, with shape or :math:`(*, C')` or :math:`(*, S, C')`.
+        """
+
+        return linear(x, self.weight, self.bias)
+
+
 class MLP(nn.Sequential):
     r"""Creates a multi-layer perceptron (MLP).
 
@@ -59,7 +142,7 @@ class MLP(nn.Sequential):
         activation: The activation function constructor. If :py:`None`, use
             :class:`torch.nn.ReLU` instead.
         normalize: Whether features are normalized between layers or not.
-        kwargs: Keyword arguments passed to :class:`torch.nn.Linear`.
+        kwargs: Keyword arguments passed to :class:`Linear`.
 
     Example:
         >>> net = MLP(64, 1, [32, 16], activation=nn.ELU)
@@ -94,13 +177,13 @@ class MLP(nn.Sequential):
             (*hidden_features, out_features),
         ):
             layers.extend([
-                nn.Linear(before, after, **kwargs),
+                Linear(before, after, **kwargs),
                 activation(),
                 normalization(),
             ])
 
         layers = layers[:-2]
-        layers = filter(lambda l: l is not None, layers)
+        layers = filter(lambda layer: layer is not None, layers)
 
         super().__init__(*layers)
 
@@ -108,6 +191,13 @@ class MLP(nn.Sequential):
         self.out_features = out_features
 
 
+class Residual(nn.Sequential):
+    r"""Creates a residual block."""
+
+    def forward(self, x: Tensor) -> Tensor:
+        return x + super().forward(x)
+
+
 class MaskedLinear(nn.Linear):
     r"""Creates a masked linear layer.
 
@@ -121,7 +211,7 @@ class MaskedLinear(nn.Linear):
     def __init__(self, adjacency: BoolTensor, **kwargs):
         super().__init__(*reversed(adjacency.shape), **kwargs)
 
-        self.register_buffer('mask', adjacency)
+        self.register_buffer("mask", adjacency)
 
     def forward(self, x: Tensor) -> Tensor:
         return F.linear(x, self.mask * self.weight, self.bias)
@@ -138,14 +228,15 @@ class MaskedMLP(nn.Sequential):
         hidden_features: The numbers of hidden features.
         activation: The activation function constructor. If :py:`None`, use
             :class:`torch.nn.ReLU` instead.
+        residual: Whether to use residual blocks or not.
 
     Example:
         >>> adjacency = torch.randn(4, 3) < 0
         >>> adjacency
-        tensor([[False,  True, False],
-                [ True, False,  True],
-                [False,  True, False],
-                [False,  True,  True]])
+        tensor([[False,  True,  True],
+                [False,  True,  True],
+                [False, False,  True],
+                [ True,  True, False]])
         >>> net = MaskedMLP(adjacency, [16, 32], activation=nn.ELU)
         >>> net
         MaskedMLP(
@@ -157,10 +248,10 @@ class MaskedMLP(nn.Sequential):
         )
         >>> x = torch.randn(3)
         >>> torch.autograd.functional.jacobian(net, x)
-        tensor([[ 0.0000,  0.0031,  0.0000],
-                [-0.0323,  0.0000, -0.0547],
-                [ 0.0000, -0.0245,  0.0000],
-                [ 0.0000,  0.0060, -0.0063]])
+        tensor([[ 0.0000, -0.0065,  0.1158],
+                [ 0.0000, -0.0089,  0.0072],
+                [ 0.0000,  0.0000,  0.0089],
+                [-0.0146, -0.0128,  0.0000]])
     """
 
     def __init__(
@@ -168,6 +259,7 @@ class MaskedMLP(nn.Sequential):
         adjacency: BoolTensor,
         hidden_features: Sequence[int] = (64, 64),
         activation: Callable[[], nn.Module] = None,
+        residual: bool = False,
     ):
         out_features, in_features = adjacency.shape
 
@@ -178,14 +270,14 @@ class MaskedMLP(nn.Sequential):
         adjacency, inverse = torch.unique(adjacency, dim=0, return_inverse=True)
 
         # P_ij = 1 if A_ik = 1 for all k such that A_jk = 1
-        precedence = adjacency.int() @ adjacency.int().t() == adjacency.sum(dim=-1)
+        precedence = adjacency.double() @ adjacency.double().t() == adjacency.sum(dim=-1)
 
         # Layers
         layers = []
 
         for i, features in enumerate((*hidden_features, out_features)):
             if i > 0:
-                mask = precedence[:, indices]
+                mask = precedence[:, indices]  # noqa: F821
             else:
                 mask = adjacency
 
@@ -199,12 +291,25 @@ class MaskedMLP(nn.Sequential):
             else:
                 mask = mask[inverse]
 
-            layers.extend([
-                MaskedLinear(adjacency=mask),
-                activation(),
-            ])
+            layers.append(MaskedLinear(adjacency=mask))
 
-        layers = layers[:-1]
+            if residual:
+                if 0 < i < len(hidden_features) and mask.shape[0] == mask.shape[1]:
+                    layers.pop()
+
+                mask = precedence[indices, :][:, indices]
+
+                layers.append(
+                    Residual(
+                        MaskedLinear(adjacency=mask),
+                        activation(),
+                        MaskedLinear(adjacency=mask),
+                    )
+                )
+            else:
+                layers.append(activation())
+
+        layers.pop()
 
         super().__init__(*layers)
 
@@ -212,18 +317,18 @@ class MaskedMLP(nn.Sequential):
         self.out_features = out_features
 
 
-class MonotonicLinear(nn.Linear):
+class MonotonicLinear(Linear):
     r"""Creates a monotonic linear layer.
 
     .. math:: y = x |W|^T + b
 
     Arguments:
-        args: Positional arguments passed to :class:`torch.nn.Linear`.
-        kwargs: Keyword arguments passed to :class:`torch.nn.Linear`.
+        args: Positional arguments passed to :class:`Linear`.
+        kwargs: Keyword arguments passed to :class:`Linear`.
     """
 
     def forward(self, x: Tensor) -> Tensor:
-        return F.linear(x, self.weight.abs(), self.bias)
+        return linear(x, self.weight.abs(), self.bias)
 
 
 class TwoWayELU(nn.ELU):
@@ -238,10 +343,13 @@ class TwoWayELU(nn.ELU):
     def forward(self, x: Tensor) -> Tensor:
         x0, x1 = torch.chunk(x, 2, dim=-1)
 
-        return torch.cat((
-            super().forward(x0),
-            -super().forward(-x1),
-        ), dim=-1)
+        return torch.cat(
+            (
+                super().forward(x0),
+                -super().forward(-x1),
+            ),
+            dim=-1,
+        )
 
 
 class MonotonicMLP(MLP):
@@ -266,18 +374,18 @@ class MonotonicMLP(MLP):
         )
         >>> x = torch.randn(3)
         >>> torch.autograd.functional.jacobian(net, x)
-        tensor([[0.8742, 0.9439, 0.9759],
-                [0.8969, 0.9716, 0.9866],
-                [1.0780, 1.1651, 1.2056],
-                [0.8596, 0.9400, 0.9502]])
+        tensor([[1.0492, 1.3094, 1.1711],
+                [1.1201, 1.3825, 1.2711],
+                [0.9397, 1.1915, 1.0787],
+                [1.1049, 1.3635, 1.2592]])
     """
 
     def __init__(self, *args, **kwargs):
-        kwargs['activation'] = TwoWayELU
-        kwargs['normalize'] = False
+        kwargs["activation"] = TwoWayELU
+        kwargs["normalize"] = False
 
         super().__init__(*args, **kwargs)
 
         for layer in self:
-            if isinstance(layer, nn.Linear):
+            if isinstance(layer, Linear):
                 layer.__class__ = MonotonicLinear
